@@ -30,6 +30,13 @@
 #define THREADING_SCHEME 8
 #endif
 
+#if ((__GNUC__==2 && defined(__GNUC_MINOR__) && __GNUC_MINOR__>=95) || (__GNUC__>2))
+#if !defined(USE_TOS) && !defined(USE_NO_TOS)
+#define USE_TOS
+#endif
+#endif
+#define MEMCMP_AS_SUBROUTINE 1
+
 #include "../generic/machine.h"
 
 /* indirect threading is faster on the Pentium, on the 486 direct
@@ -55,9 +62,17 @@
 #define PFA(cfa)	(((Cell *)cfa)+2)
 /* PFA1 is a special version for use just after a NEXT1 */
 #define PFA1(cfa)	PFA(cfa)
+/* a special version of CODE_ADDRESS for DOES_CODE1 and DOES_CODE */
+#define CODE_ADDRESS1(cfa) \
+    ({long _cfa1 = (long)(cfa); (Label)(_cfa1+*((long *)(_cfa1+1))+5);})
+
 /* CODE_ADDRESS is the address of the code jumped to through the code field */
 #define CODE_ADDRESS(cfa) \
-    ({long _cfa = (long)(cfa); (Label)(_cfa+*((long *)(_cfa+1))+5);})
+    ({long _cfa = (long)(cfa); \
+       (((*(unsigned char *)_cfa)==CALL) ? \
+       CODE_ADDRESS1(_cfa) : \
+       (Label)(_cfa));})
+
 /* MAKE_CF creates an appropriate code field at the cfa; ca is the code address */
 #define MAKE_CF(cfa,ca)	({long _cfa = (long)(cfa); \
                           long _ca  = (long)(ca); \
@@ -71,11 +86,11 @@
    long _ca = (long)(CODE_ADDRESS(_xt)); \
    ((((*(unsigned char *)_xt) == CALL) \
      && ((*(unsigned char *)_ca) == JMP) \
-     && ((long)(CODE_ADDRESS(_ca)) == (long)&&dodoes)) \
+     && ((long)(CODE_ADDRESS1(_ca)) == (long)&&dodoes)) \
     ? _ca+DOES_HANDLER_SIZE : 0L); })
 
 /* this is a special version of DOES_CODE for use in dodoes */
-#define DOES_CODE1(label)      (CODE_ADDRESS(label)+DOES_HANDLER_SIZE)
+#define DOES_CODE1(label)      (CODE_ADDRESS1(label)+DOES_HANDLER_SIZE)
 
 /* this stores a jump dodoes at addr */
 #define MAKE_DOES_CF(addr,doesp)	({long _addr = (long)(addr); \
@@ -104,9 +119,21 @@
 /* this works with 2.6.3 (and quite well, too) */
 /* since this is not very demanding, it's the default for other gcc versions */
 #if defined(USE_TOS) && !defined(CFA_NEXT)
+#if ((__GNUC__==2 && defined(__GNUC_MINOR__) && __GNUC_MINOR__>=95) || (__GNUC__>2))
+     /* gcc 2.95 has a better register allocater */
+#define SPREG asm("%esi")
+#define RPREG asm("%edi")
+#define TOSREG asm("%ebx")
+#else
 #define IPREG asm("%ebx")
+#endif
+#else
+#if ((__GNUC__==2 && defined(__GNUC_MINOR__) && __GNUC_MINOR__>=95) || (__GNUC__>2))
+#define SPREG asm("%esi")
+#define RPREG asm("%edi")
 #else
 #define SPREG asm("%ebx")
+#endif
 #endif /* USE_TOS && !CFA_NEXT */
 #endif /* gcc-version */
 #endif /* FORCE_REG */
