@@ -38,6 +38,12 @@
 #include <sys/time.h>
 #include <unistd.h>
 #include <pwd.h>
+#include <dirent.h>
+#ifdef HAVE_FNMATCH_H
+#include <fnmatch.h>
+#else
+#include "fnmatch.h"
+#endif
 #else
 #include "systypes.h"
 #endif
@@ -80,6 +86,11 @@ int emitcounter;
 #endif
 #define NULLC '\0'
 
+#ifdef MEMCMP_AS_SUBROUTINE
+extern int gforth_memcmp(const char * s1, const char * s2, size_t n);
+#define memcmp(s1,s2,n) gforth_memcmp(s1,s2,n)
+#endif
+
 #ifdef HAS_FILE
 char *cstr(Char *from, UCell size, int clear)
 /* return a C-string corresponding to the Forth string ( FROM SIZE ).
@@ -165,7 +176,7 @@ char *tilde_cstr(Char *from, UCell size, int clear)
 #endif
 
 #ifdef HAS_FILE
-static char* fileattr[6]={"r","rb","r+","r+b","w","wb"};
+static char* fileattr[6]={"rb","rb","r+b","r+b","wb","wb"};
 
 #ifndef O_BINARY
 #define O_BINARY 0
@@ -175,9 +186,9 @@ static char* fileattr[6]={"r","rb","r+","r+b","w","wb"};
 #endif
 
 static int ufileattr[6]= {
-  O_RDONLY|O_TEXT, O_RDONLY|O_BINARY,
-  O_RDWR  |O_TEXT, O_RDWR  |O_BINARY,
-  O_WRONLY|O_TEXT, O_WRONLY|O_BINARY };
+  O_RDONLY|O_BINARY, O_RDONLY|O_BINARY,
+  O_RDWR  |O_BINARY, O_RDWR  |O_BINARY,
+  O_WRONLY|O_BINARY, O_WRONLY|O_BINARY };
 #endif
 
 /* if machine.h has not defined explicit registers, define them as implicit */
