@@ -28,7 +28,9 @@
 #include <string.h>
 #include <math.h>
 #include <sys/types.h>
+#ifndef STANDALONE
 #include <sys/stat.h>
+#endif
 #include <fcntl.h>
 #include <assert.h>
 #include <stdlib.h>
@@ -78,9 +80,24 @@ char *progname;
 char *progname = "gforth";
 int optind = 1;
 #endif
+
+#ifdef HAS_DEBUG
 static int debug=0;
+#else
+# define debug 0
+# define perror(x...)
+# define fprintf(x...)
+#endif
+
 ImageHeader *gforth_header;
 
+#ifdef MEMCMP_AS_SUBROUTINE
+int gforth_memcmp(const char * s1, const char * s2, size_t n)
+{
+  return memcmp(s1, s2, n);
+}
+#endif
+
 /* image file format:
  *  "#! binary-path -i\n" (e.g., "#! /usr/local/bin/gforth-0.4.0 -i\n")
  *   padding to a multiple of 8
@@ -113,15 +130,16 @@ void relocate(Cell *image, const char *bitstring, int size, Label symbols[])
   int i=0, j, k, steps=(size/sizeof(Cell))/RELINFOBITS;
   Cell token;
   char bits;
-/*   static char bits[8]={0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};*/
 
 /*  printf("relocating %x[%x]\n", image, size); */
    
-  for(k=0; k<=steps; k++)
+  size/=sizeof(Cell);
+
+  for(k=0; k<=steps; k++) {
     for(j=0, bits=bitstring[k]; j<RELINFOBITS; j++, i++, bits<<=1) {
       /*      fprintf(stderr,"relocate: image[%d]\n", i);*/
-      if(bits & (1U << (RELINFOBITS-1))) {
-	/* fprintf(stderr,"relocate: image[%d]=%d\n", i, image[i]);*/
+      if((i < size) && (bits & (1U << (RELINFOBITS-1)))) {
+	/* fprintf(stderr,"relocate: image[%d]=%d of %d\n", i, image[i], size/sizeof(Cell)); */
 	if((token=image[i])<0)
 	  switch(token)
 	    {
@@ -136,7 +154,7 @@ void relocate(Cell *image, const char *bitstring, int size, Label symbols[])
 	    case CF(DOESJUMP): MAKE_DOES_HANDLER(image+i); break;
 #endif /* !defined(DOUBLY_INDIRECT) */
 	    case CF(DODOES)  :
-	      MAKE_DOES_CF(image+i,image[i+1]+((Cell)image));
+	      MAKE_DOES_CF(image+i,(Xt *)(image[i+1]+((Cell)image)));
 	      break;
 	    default          :
 /*	      printf("Code field generation image[%x]:=CA(%x)\n",
@@ -147,6 +165,7 @@ void relocate(Cell *image, const char *bitstring, int size, Label symbols[])
 	  image[i]+=(Cell)image;
       }
     }
+  }
   ((ImageHeader*)(image))->base = (Address) image;
 }
 
@@ -188,31 +207,48 @@ Address verbose_malloc(Cell size)
   return r;
 }
 
-Address my_alloc(Cell size)
+static Address next_address=0;
+void after_alloc(Address r, Cell size)
 {
-#if HAVE_MMAP
-  static Address next_address=0;
-  Address r;
+  if (r != (Address)-1) {
+    if (debug)
+      fprintf(stderr, "success, address=$%lx\n", (long) r);
+    if (pagesize != 1)
+      next_address = (Address)(((((Cell)r)+size-1)&-pagesize)+2*pagesize); /* leave one page unmapped */
+  } else {
+    if (debug)
+      fprintf(stderr, "failed: %s\n", strerror(errno));
+  }
+}
 
-#if defined(MAP_ANON)
-  if (debug)
-    fprintf(stderr,"try mmap($%lx, $%lx, ..., MAP_ANON, ...); ", (long)next_address, (long)size);
-  r=mmap(next_address, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE, -1, 0);
-#else /* !defined(MAP_ANON) */
-  /* Ultrix (at least) does not define MAP_FILE and MAP_PRIVATE (both are
-     apparently defaults) */
+#ifndef MAP_FAILED
+#define MAP_FAILED ((Address) -1)
+#endif
 #ifndef MAP_FILE
 # define MAP_FILE 0
 #endif
 #ifndef MAP_PRIVATE
 # define MAP_PRIVATE 0
 #endif
+
+#if defined(HAVE_MMAP)
+static Address alloc_mmap(Cell size)
+{
+  Address r;
+
+#if defined(MAP_ANON)
+  if (debug)
+    fprintf(stderr,"try mmap($%lx, $%lx, ..., MAP_ANON, ...); ", (long)next_address, (long)size);
+  r = mmap(next_address, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE, -1, 0);
+#else /* !defined(MAP_ANON) */
+  /* Ultrix (at least) does not define MAP_FILE and MAP_PRIVATE (both are
+     apparently defaults) */
   static int dev_zero=-1;
 
   if (dev_zero == -1)
     dev_zero = open("/dev/zero", O_RDONLY);
   if (dev_zero == -1) {
-    r = (Address)-1;
+    r = MAP_FAILED;
     if (debug)
       fprintf(stderr, "open(\"/dev/zero\"...) failed (%s), no mmap; ", 
 	      strerror(errno));
@@ -222,16 +258,19 @@ Address my_alloc(Cell size)
     r=mmap(next_address, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_FILE|MAP_PRIVATE, dev_zero, 0);
   }
 #endif /* !defined(MAP_ANON) */
+  after_alloc(r, size);
+  return r;  
+}
+#endif
 
-  if (r != (Address)-1) {
-    if (debug)
-      fprintf(stderr, "success, address=$%lx\n", (long) r);
-    if (pagesize != 1)
-      next_address = (Address)(((((Cell)r)+size-1)&-pagesize)+2*pagesize); /* leave one page unmapped */
+Address my_alloc(Cell size)
+{
+#if HAVE_MMAP
+  Address r;
+
+  r=alloc_mmap(size);
+  if (r!=MAP_FAILED)
     return r;
-  }
-  if (debug)
-    fprintf(stderr, "failed: %s\n", strerror(errno));
 #endif /* HAVE_MMAP */
   /* use malloc as fallback */
   return verbose_malloc(size);
@@ -240,11 +279,33 @@ Address my_alloc(Cell size)
 #if (defined(mips) && !defined(INDIRECT_THREADED))
 /* the 256MB jump restriction on the MIPS architecture makes the
    combination of direct threading and mmap unsafe. */
+#define mips_dict_alloc 1
 #define dict_alloc(size) verbose_malloc(size)
 #else
 #define dict_alloc(size) my_alloc(size)
 #endif
 
+Address dict_alloc_read(FILE *file, Cell imagesize, Cell dictsize, Cell offset)
+{
+  Address image = MAP_FAILED;
+
+#if defined(HAVE_MMAP) && !defined(mips_dict_alloc)
+  if (offset==0) {
+    image=alloc_mmap(dictsize);
+    if (debug)
+      fprintf(stderr,"try mmap($%lx, $%lx, ..., MAP_FIXED|MAP_FILE, imagefile, 0); ", (long)image, (long)imagesize);
+    image = mmap(image, imagesize, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_FIXED|MAP_FILE|MAP_PRIVATE, fileno(file), 0);
+    after_alloc(image,dictsize);
+  }
+#endif /* defined(MAP_ANON) && !defined(mips_dict_alloc) */
+  if (image == MAP_FAILED) {
+    image = dict_alloc(dictsize+offset)+offset;
+    rewind(file);  /* fseek(imagefile,0L,SEEK_SET); */
+    fread(image, 1, imagesize, file);
+  }
+  return image;
+}
+
 void set_stack_sizes(ImageHeader * header)
 {
   if (dictsize==0)
@@ -297,9 +358,9 @@ int go_forth(Address image, int stack, Cell *entries)
   for(;stack>0;stack--)
     *--sp0=entries[stack-1];
 
+#ifdef SYSSIGNALS
   get_winsize();
    
-#ifdef SYSSIGNALS
   install_signal_handlers(); /* right place? */
   
   if ((throw_code=setjmp(throw_jmp_buf))) {
@@ -320,15 +381,16 @@ int go_forth(Address image, int stack, Cell *entries)
     rp0 = signal_return_stack+8;
     /* fprintf(stderr, "rp=$%x\n",rp0);*/
     
-    return((int)engine(image_header->throw_entry, signal_data_stack+7,
+    return((int)(Cell)engine(image_header->throw_entry, signal_data_stack+7,
 		       rp0, signal_fp_stack, 0));
   }
 #endif
 
-  return((int)engine(ip0,sp0,rp0,fp0,lp0));
+  return((int)(Cell)engine(ip0,sp0,rp0,fp0,lp0));
 }
 
 
+#ifndef INCLUDE_IMAGE
 void print_sizes(Cell sizebyte)
      /* print size information */
 {
@@ -341,7 +403,6 @@ void print_sizes(Cell sizebyte)
 	  1 << ((sizebyte >> 5) & 3));
 }
 
-#ifndef INCLUDE_IMAGE
 Address loader(FILE *imagefile, char* filename)
 /* returns the address of the image proper (after the preamble) */
 {
@@ -415,15 +476,15 @@ Address loader(FILE *imagefile, char* filename)
   if (debug)
     fprintf(stderr,"pagesize=%ld\n",(unsigned long) pagesize);
 
-  image = dict_alloc(preamblesize+dictsize+data_offset)+data_offset;
-  rewind(imagefile);  /* fseek(imagefile,0L,SEEK_SET); */
-  if (clear_dictionary)
-    memset(image, 0, dictsize);
-  fread(image, 1, preamblesize+header.image_size, imagefile);
+  image = dict_alloc_read(imagefile, preamblesize+header.image_size,
+			  preamblesize+dictsize, data_offset);
   imp=image+preamblesize;
+  if (clear_dictionary)
+    memset(imp+header.image_size, 0, dictsize-header.image_size);
   if(header.base==0) {
     Cell reloc_size=((header.image_size-1)/sizeof(Cell))/8+1;
     char reloc_bits[reloc_size];
+    fseek(imagefile, preamblesize+header.image_size, SEEK_SET);
     fread(reloc_bits, 1, reloc_size, imagefile);
     relocate((Cell *)imp, reloc_bits, header.image_size, symbols);
 #if 0
@@ -568,6 +629,7 @@ void gforth_args(int argc, char ** argv, char ** path, char ** imagename)
   while (1) {
     int option_index=0;
     static struct option opts[] = {
+      {"appl-image", required_argument, NULL, 'a'},
       {"image-file", required_argument, NULL, 'i'},
       {"dictionary-size", required_argument, NULL, 'm'},
       {"data-stack-size", required_argument, NULL, 'd'},
@@ -589,13 +651,10 @@ void gforth_args(int argc, char ** argv, char ** path, char ** imagename)
     
     c = getopt_long(argc, argv, "+i:m:d:r:f:l:p:vh", opts, &option_index);
     
-    if (c==EOF)
-      break;
-    if (c=='?') {
-      optind--;
-      break;
-    }
     switch (c) {
+    case EOF: return;
+    case '?': optind--; return;
+    case 'a': *imagename = optarg; return;
     case 'i': *imagename = optarg; break;
     case 'm': dictsize = convsize(optarg,sizeof(Cell)); break;
     case 'd': dsize = convsize(optarg,sizeof(Cell)); break;
@@ -605,8 +664,9 @@ void gforth_args(int argc, char ** argv, char ** path, char ** imagename)
     case 'p': *path = optarg; break;
     case 'v': fprintf(stderr, "gforth %s\n", VERSION); exit(0);
     case 'h': 
-      fprintf(stderr, "Usage: %s [engine options] [image arguments]\n\
+      fprintf(stderr, "Usage: %s [engine options] ['--'] [image arguments]\n\
 Engine Options:\n\
+  --appl-image FILE		    equivalent to '--image-file=FILE --'\n\
   --clear-dictionary		    Initialize the dictionary with 0 bytes\n\
   -d SIZE, --data-stack-size=SIZE   Specify data stack size\n\
   --debug			    Print debugging information during startup\n\
@@ -626,7 +686,6 @@ SIZE arguments consist of an integer followed by a unit. The unit can be\n\
 	      argv[0]);
       optind--;
       return;
-      exit(0);
     }
   }
 }
@@ -639,7 +698,11 @@ extern const char reloc_bits[];
 
 int main(int argc, char **argv, char **env)
 {
+#ifdef HAS_OS
   char *path = getenv("GFORTHPATH") ? : DEFAULTPATH;
+#else
+  char *path = DEFAULTPATH;
+#endif
 #ifndef INCLUDE_IMAGE
   char *imagename="gforth.fi";
   FILE *image_file;
